/**
 * 
 */
package org.springblade.etl.connector.sink;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.util.ArrayUtil;
import cn.hutool.core.util.StrUtil;
import io.swagger.annotations.ApiModelProperty;
import lombok.Data;
import lombok.EqualsAndHashCode;
import org.springblade.core.tool.exception.ServiceException;
import org.springblade.datasource.context.DialectContext;
import org.springblade.etl.connector.base.SeatunnelConnector;
import org.springblade.etl.connector.sink.config.PlatomixJdbcSinkBaseConfig;
import org.springblade.etl.connector.sink.config.SeatunnelJdbcSinkConfig;
import org.springblade.etl.entity.bi.ReportDatasource;
import org.springblade.etl.enums.JdbcType;
import org.springblade.etl.service.bi.ReportDatasourceService;
import org.springblade.etl.util.EtlJsonUtil;
import org.springblade.etl.util.ValidateUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * @author wookvn
 *
 */
@Component
public class PlatomixJdbcSink implements SeatunnelConnector<SeatunnelJdbcSinkConfig> {

	@Autowired
	private ReportDatasourceService reportDatasourceService;
	
	@Autowired
	private CommonJdbcSink commonJdbcSink;
	
	@Autowired
	private DialectContext dialectContext;
	
	@Data
	@EqualsAndHashCode(callSuper = true)
	public static class Config extends PlatomixJdbcSinkBaseConfig {
		/**
		 */
		private static final long serialVersionUID = 1L;
		
		@ApiModelProperty("source的主键列表")
		private List<String> primaryKeys;
		@ApiModelProperty("表字段列表【columnFields和primaryKeys必须有一个】")
		private List<String> columnFields;
	}

	@Override
	public SeatunnelJdbcSinkConfig buildConfig(String configJson) {
		Config ptConfig = (Config)validate(configJson);
		ReportDatasource dsModel = getDs(ptConfig.getDatasourceId());
		SeatunnelJdbcSinkConfig seatunnelConfig = new SeatunnelJdbcSinkConfig();
		seatunnelConfig.setJdbcType(JdbcType.ofDsType(dsModel.getDsType()));
		seatunnelConfig.setUrl(dsModel.getUrl());
		seatunnelConfig.setUsername(dsModel.getDsUser());
		seatunnelConfig.setPassword(dsModel.getDsPassword());
		seatunnelConfig.setDriver(dsModel.getDriverClassName());
		String database = dialectContext.getDialect(dsModel.getDsTypeSyl()).dbSchema(dsModel);
		seatunnelConfig.setDatabase(database);
		seatunnelConfig.setTable(ptConfig.getTableName());
		String querySql = null;
		//@TODO 如果主键和字段都存在，获取upsert语句 【以后再做处理】
		if(CollectionUtil.isNotEmpty(ptConfig.getPrimaryKeys()) && CollectionUtil.isNotEmpty(ptConfig.getColumnFields())) {
//			querySql = dialectContext.getDialect(dsModel.getDsTypeSyl()).getUpsertStatement(dsModel, ptConfig.getTableName(), ptConfig.getColumnFields().toArray(String[]::new), ptConfig.getPrimaryKeys().toArray(String[]::new));
		}
		//upsert语句获取失败，并且源库表主键也不存在，生成insert语句
		if(StrUtil.isBlank(querySql) && CollectionUtil.isEmpty(ptConfig.getPrimaryKeys())) {
			String tableName = dialectContext.getDialect(dsModel.getDsTypeSyl()).dbTableName(dsModel, ptConfig.getTableName());
			querySql = String.format("insert into %s(%s) values(%s)",
					tableName, String.join(",", ptConfig.getColumnFields()),
					IntStream.range(0, ptConfig.getColumnFields().size()).mapToObj(t -> "?").collect(Collectors.joining(",")));
		}
		if(StrUtil.isBlank(querySql)){
			seatunnelConfig.setPrimary_keys(ptConfig.getPrimaryKeys());
			seatunnelConfig.setSupport_upsert_by_query_primary_key_exist(CollectionUtil.isNotEmpty(ptConfig.getPrimaryKeys()));
		}
		seatunnelConfig.setQuery(querySql);
		return seatunnelConfig;
	}
	
	@Override
	public Object validate(String configJson) {
		Config ptConfig = EtlJsonUtil.parseJsonString(configJson, PlatomixJdbcSink.Config.class);
		ValidateUtil.validate(ptConfig);
		String[] primaryKeyArr = Optional.ofNullable(ptConfig.getPrimaryKeys()).orElse(new ArrayList<>()).toArray(String[]::new);
		String[] getColumnFieldArr = Optional.ofNullable(ptConfig.getColumnFields()).orElse(new ArrayList<>()).toArray(String[]::new);
		if(ArrayUtil.isAllNull(primaryKeyArr) && ArrayUtil.isAllNull(getColumnFieldArr)) {
			throw new ServiceException("输出源主键或字段列表必须有一个非空");
		}
		//清除所有NULL值
		if(ptConfig.getPrimaryKeys() != null) {
			ptConfig.getPrimaryKeys().removeIf(item -> item == null);
		}
		if(ptConfig.getColumnFields() != null) {
			ptConfig.getColumnFields().removeIf(item -> item == null);
		}
		return ptConfig;
	}

	@Override
	public void prepare(String configJson) {
		commonJdbcSink.prepare(configJson);
	}
	
	/**
	 * 获取数据源
	 * @param databaseId
	 * @return
	 */
	private ReportDatasource getDs(Long databaseId) {
		ReportDatasource sourceModel = reportDatasourceService.getById(databaseId);
		if(sourceModel == null) {
			throw new ServiceException("未查询到数据源");
		}
		return sourceModel;
	}

//	String getInsertIntoStatement(String database, String tableName, String[] fieldNames) {
//		String columns =
//				Arrays.stream(fieldNames)
//						.map(this::quoteIdentifier)
//						.collect(Collectors.joining(", "));
//		String placeholders =
//				Arrays.stream(fieldNames)
//						.map(fieldName -> ":" + fieldName)
//						.collect(Collectors.joining(", "));
//		return String.format(
//				"INSERT INTO %s (%s) VALUES (%s)",
//				tableIdentifier(database, tableName), columns, placeholders);
//	}
//
//	String tableIdentifier(String database, String tableName) {
//		return quoteIdentifier(database) + "." + quoteIdentifier(tableName);
//	}
//
//	public String quoteIdentifier(String identifier) {
//		return "`" + identifier + "`";
//	}
//
//	public Optional<String> getUpsertStatement(
//			String database, String tableName, String[] fieldNames, String[] uniqueKeyFields) {
//		String updateClause =
//				Arrays.stream(fieldNames)
//						.map(
//								fieldName ->
//										quoteIdentifier(fieldName)
//												+ "=VALUES("
//												+ quoteIdentifier(fieldName)
//												+ ")")
//						.collect(Collectors.joining(", "));
//		String upsertSQL =
//				getInsertIntoStatement(database, tableName, fieldNames)
//						+ " ON DUPLICATE KEY UPDATE "
//						+ updateClause;
//		return Optional.of(upsertSQL);
//	}
//
//	public Optional<String> getOracleUpsertStatement(
//			String database, String tableName, String[] fieldNames, String[] uniqueKeyFields) {
//		List<String> nonUniqueKeyFields =
//				Arrays.stream(fieldNames)
//						.filter(fieldName -> !Arrays.asList(uniqueKeyFields).contains(fieldName))
//						.collect(Collectors.toList());
//		String valuesBinding =
//				Arrays.stream(fieldNames)
//						.map(fieldName -> ":" + fieldName + " " + quoteIdentifier(fieldName))
//						.collect(Collectors.joining(", "));
//
//		String usingClause = String.format("SELECT %s FROM DUAL", valuesBinding);
//		String onConditions =
//				Arrays.stream(uniqueKeyFields)
//						.map(
//								fieldName ->
//										String.format(
//												"TARGET.%s=SOURCE.%s",
//												quoteIdentifier(fieldName),
//												quoteIdentifier(fieldName)))
//						.collect(Collectors.joining(" AND "));
//		String updateSetClause =
//				nonUniqueKeyFields.stream()
//						.map(
//								fieldName ->
//										String.format(
//												"TARGET.%s=SOURCE.%s",
//												quoteIdentifier(fieldName),
//												quoteIdentifier(fieldName)))
//						.collect(Collectors.joining(", "));
//		String insertFields =
//				Arrays.stream(fieldNames)
//						.map(this::quoteIdentifier)
//						.collect(Collectors.joining(", "));
//		String insertValues =
//				Arrays.stream(fieldNames)
//						.map(fieldName -> "SOURCE." + quoteIdentifier(fieldName))
//						.collect(Collectors.joining(", "));
//
//		String upsertSQL =
//				String.format(
//						" MERGE INTO %s TARGET"
//								+ " USING (%s) SOURCE"
//								+ " ON (%s) "
//								+ " WHEN MATCHED THEN"
//								+ " UPDATE SET %s"
//								+ " WHEN NOT MATCHED THEN"
//								+ " INSERT (%s) VALUES (%s)",
//						tableIdentifier(database, tableName),
//						usingClause,
//						onConditions,
//						updateSetClause,
//						insertFields,
//						insertValues);
//
//		return Optional.of(upsertSQL);
//	}
//
//	public Optional<String> getSqlserverUpsertStatement(
//			String database, String tableName, String[] fieldNames, String[] uniqueKeyFields) {
//		List<String> nonUniqueKeyFields =
//				Arrays.stream(fieldNames)
//						.filter(fieldName -> !Arrays.asList(uniqueKeyFields).contains(fieldName))
//						.collect(Collectors.toList());
//		String valuesBinding =
//				Arrays.stream(fieldNames)
//						.map(fieldName -> ":" + fieldName + " " + quoteIdentifier(fieldName))
//						.collect(Collectors.joining(", "));
//
//		String usingClause = String.format("SELECT %s", valuesBinding);
//		String onConditions =
//				Arrays.stream(uniqueKeyFields)
//						.map(
//								fieldName ->
//										String.format(
//												"[TARGET].%s=[SOURCE].%s",
//												quoteIdentifier(fieldName),
//												quoteIdentifier(fieldName)))
//						.collect(Collectors.joining(" AND "));
//		String updateSetClause =
//				nonUniqueKeyFields.stream()
//						.map(
//								fieldName ->
//										String.format(
//												"[TARGET].%s=[SOURCE].%s",
//												quoteIdentifier(fieldName),
//												quoteIdentifier(fieldName)))
//						.collect(Collectors.joining(", "));
//		String insertFields =
//				Arrays.stream(fieldNames)
//						.map(this::quoteIdentifier)
//						.collect(Collectors.joining(", "));
//		String insertValues =
//				Arrays.stream(fieldNames)
//						.map(fieldName -> "[SOURCE]." + quoteIdentifier(fieldName))
//						.collect(Collectors.joining(", "));
//		String upsertSQL =
//				String.format(
//						"MERGE INTO %s.%s AS [TARGET]"
//								+ " USING (%s) AS [SOURCE]"
//								+ " ON (%s)"
//								+ " WHEN MATCHED THEN"
//								+ " UPDATE SET %s"
//								+ " WHEN NOT MATCHED THEN"
//								+ " INSERT (%s) VALUES (%s);",
//						database,
//						tableName,
//						usingClause,
//						onConditions,
//						updateSetClause,
//						insertFields,
//						insertValues);
//
//		return Optional.of(upsertSQL);
//	}
//
//	public Optional<String> getPostgreUpsertStatement(
//			String database, String tableName, String[] fieldNames, String[] uniqueKeyFields) {
//		String uniqueColumns =
//				Arrays.stream(uniqueKeyFields)
//						.map(this::quoteIdentifier)
//						.collect(Collectors.joining(", "));
//		String updateClause =
//				Arrays.stream(fieldNames)
//						.map(
//								fieldName ->
//										quoteIdentifier(fieldName)
//												+ "=EXCLUDED."
//												+ quoteIdentifier(fieldName))
//						.collect(Collectors.joining(", "));
//		String upsertSQL =
//				String.format(
//						"%s ON CONFLICT (%s) DO UPDATE SET %s",
//						getInsertIntoStatement(database, tableName, fieldNames),
//						uniqueColumns,
//						updateClause);
//		return Optional.of(upsertSQL);
//	}
//
//	public static void main(String[] args) {
//		PlatomixJdbcSink platomixJdbcSink = new PlatomixJdbcSink();
//		String insertIntoStatement = platomixJdbcSink.getInsertIntoStatement("testdb", "table01", new String[]{"id", "name"});
//		System.out.println(insertIntoStatement);
//		Optional<String> upsertStatement = platomixJdbcSink.getUpsertStatement("testdb", "table01", new String[]{"id", "name"}, new String[]{"id"});
//		System.out.println(upsertStatement.get());
//		Optional<String> oracleUpsertStatement = platomixJdbcSink.getOracleUpsertStatement("testdb", "table01", new String[]{"id", "name"}, new String[]{"id"});
//		System.out.println(oracleUpsertStatement.get());
//		Optional<String> sqlserverUpsertStatement = platomixJdbcSink.getSqlserverUpsertStatement("testdb", "table01", new String[]{"id", "name"}, new String[]{"id"});
//		System.out.println(sqlserverUpsertStatement.get());
//		Optional<String> psqlUpsertStatement = platomixJdbcSink.getPostgreUpsertStatement("testdb", "table01", new String[]{"id", "name"}, new String[]{"id"});
//		System.out.println(psqlUpsertStatement.get());
//	}
}
