/*
 * Copyright (c) 2018-2028, aokunsang All rights reserved.
 * 我问青山何时老，青山问我几时闲，
 * 不是闲人闲不得，能闲必非等闲人。
 */
package org.springblade.etl.connector.source;

import cn.hutool.core.collection.CollectionUtil;
import io.swagger.annotations.ApiModelProperty;
import lombok.Data;
import org.springblade.core.tool.exception.ServiceException;
import org.springblade.datasource.dialect.MongoGmetryDialect;
import org.springblade.datasource.model.TableColumnDto;
import org.springblade.etl.connector.base.SeatunnelConfig;
import org.springblade.etl.connector.base.SeatunnelConnector;
import org.springblade.etl.connector.source.config.SeatunnelMongoSourceConfig;
import org.springblade.etl.connector.source.enums.FieldType;
import org.springblade.etl.entity.bi.ReportDatasource;
import org.springblade.etl.service.bi.ReportDatasourceService;
import org.springblade.etl.util.EtlJsonUtil;
import org.springblade.etl.util.ValidateUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.validation.constraints.NotEmpty;
import javax.validation.constraints.NotNull;
import java.math.BigDecimal;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * mongo 输入端数据源
 *
 * https://seatunnel.apache.org/docs/2.3.8/connector-v2/source/MongoDB
 *
 * @author wookvn
 * @datetime 2024-10-29 17:43:14
 */
@Component
public class PlatomixMongoSource implements SeatunnelConnector<SeatunnelMongoSourceConfig> {

    @Autowired
    private ReportDatasourceService reportDatasourceService;

    @Autowired
    private MongoGmetryDialect mongoGmetryDialect;

    final static Map<String, FieldType> SEATUNNEL_MONGO_TYPE_MAP = new HashMap<>();
    static {
        SEATUNNEL_MONGO_TYPE_MAP.put("OBJECT_ID", FieldType.STRING);
        SEATUNNEL_MONGO_TYPE_MAP.put("STRING", FieldType.STRING);
        SEATUNNEL_MONGO_TYPE_MAP.put("BINARY", FieldType.STRING);
        SEATUNNEL_MONGO_TYPE_MAP.put("ARRAY", FieldType.STRING);
        SEATUNNEL_MONGO_TYPE_MAP.put("DOCUMENT", FieldType.STRING);
        SEATUNNEL_MONGO_TYPE_MAP.put("OBJECT", FieldType.STRING);
        SEATUNNEL_MONGO_TYPE_MAP.put("INT32", FieldType.INT);
        SEATUNNEL_MONGO_TYPE_MAP.put("INT64", FieldType.BIGINT);
        SEATUNNEL_MONGO_TYPE_MAP.put("DOUBLE", FieldType.DOUBLE);
        SEATUNNEL_MONGO_TYPE_MAP.put("DECIMAL128", FieldType.DECIMAL);
        SEATUNNEL_MONGO_TYPE_MAP.put("DATE", FieldType.DATE);
        SEATUNNEL_MONGO_TYPE_MAP.put("DATE_TIME", FieldType.TIMESTAMP);
        SEATUNNEL_MONGO_TYPE_MAP.put("TIMESTAMP", FieldType.TIMESTAMP);
    }

    @Data
    public static class Config {
        @NotNull(message = "输入源不能为空")
        @ApiModelProperty("数据源ID")
        private Long datasourceId;
        @NotEmpty(message = "输入源表不能为空")
        @ApiModelProperty("数据表")
        private String tableName;
        @ApiModelProperty("数据字段列表")
        @NotEmpty(message = "输入源数据字段列表不能为空")
        private List<String> columnFields;
    }

    @Override
    public SeatunnelMongoSourceConfig buildConfig(String configJson) {
        PlatomixMongoSource.Config ptConfig = (PlatomixMongoSource.Config)validate(configJson);
        ReportDatasource sourceModel = reportDatasourceService.getById(ptConfig.getDatasourceId());
        if(sourceModel == null) {
            throw new ServiceException("未查询到数据源|datasourceId="+ptConfig.getDatasourceId());
        }
        SeatunnelMongoSourceConfig seatunnelConfig = new SeatunnelMongoSourceConfig();
        seatunnelConfig.setUri(sourceModel.getUrl());
        seatunnelConfig.setDatabase(sourceModel.getDsDb());
        seatunnelConfig.setCollection(ptConfig.getTableName());
        Map<String, String> fieldMap = new LinkedHashMap<>();

        //获取字段原始类型
        List<TableColumnDto> tableColumnDtos = mongoGmetryDialect.listTableColumnsByTableViewName(sourceModel, ptConfig.getTableName());
        Map<String, String> physicsColumnTypeMap = tableColumnDtos.stream().collect(Collectors.toMap(TableColumnDto::getColumnName, TableColumnDto::getDataType));

        //物理表字段类型转换为seatunnel的字段类型
        ptConfig.getColumnFields().stream().forEach(fieldName -> {
            String physicsColumnType = physicsColumnTypeMap.getOrDefault(fieldName, "__PLATOMIX_UNKNON__");
            FieldType fieldType = SEATUNNEL_MONGO_TYPE_MAP.getOrDefault(physicsColumnType.toUpperCase(), FieldType.STRING);
            fieldMap.put(fieldName, fieldType == FieldType.DECIMAL ? "DECIMAL(30, 2)" : fieldType.name());
        });
        seatunnelConfig.setSchema(new SeatunnelConfig.MongoSchemaField(fieldMap));
        return seatunnelConfig;
    }

    @Override
    public Object validate(String configJson) {
        PlatomixMongoSource.Config ptConfig = EtlJsonUtil.parseJsonString(configJson, PlatomixMongoSource.Config.class);
        ValidateUtil.validate(ptConfig);
        return ptConfig;
    }
}
