/**
 * 
 */
package org.springblade.etl.task.init;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springblade.core.tool.utils.CollectionUtil;
import org.springblade.core.tool.utils.SpringUtil;
import org.springblade.etl.engine.datax.event.JobRunEvent;
import org.springblade.etl.engine.datax.event.JobRunInstance;
import org.springblade.etl.engine.datax.event.JobRunTarget;
import org.springblade.etl.entity.DsyJobInstance;
import org.springblade.etl.entity.JobInstance;
import org.springblade.etl.mapper.DynamicTableMapper;
import org.springblade.etl.mapper.PlatomixSyncLockMapper;
import org.springblade.etl.service.DsyJobInstanceService;
import org.springblade.etl.service.JobInstanceService;
import org.springblade.etl.task.model.DaemonTaskConfig;
import org.springblade.etl.task.model.DaemonTaskType;
import org.springblade.etl.task.thread.DaemonFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.core.annotation.Order;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @author wookvn
 *
 */
@Slf4j
@Component
@Order(value = 1)
@RequiredArgsConstructor
@ConditionalOnProperty(prefix = "platomix.etl", value = "enabled", havingValue = "true")
public class SystemInit implements ApplicationRunner, DisposableBean {

	private final JobInstanceService jobInstanceService;
	
	private final DsyJobInstanceService dsyJobInstanceService;

//	private final StringRedisTemplate stringRedisTemplate;

	final static String ETL_INIT_LOCK_KEY = "etllock";

	private final PlatomixSyncLockMapper platomixSyncLockMapper;

	final static Integer ETL_INIT_LOCK_MIN = 1;

	@Override
	public void run(ApplicationArguments args) throws Exception {
		//任务线程池先启动
		DaemonFactory.start();
		long lockTs = System.currentTimeMillis();
		long lockTsBack = lockTs-ETL_INIT_LOCK_MIN * 60 * 1000;
		int i = platomixSyncLockMapper.updateSyncLock(ETL_INIT_LOCK_KEY,lockTs,lockTsBack);
		if (i > 0) {
			//作业监控初始化
			List<DaemonTaskConfig> allTaskConfig = new ArrayList<>();
			allTaskConfig.addAll(initEtlTaskMonitor());
			allTaskConfig.addAll(initDsyRunningInstanceMonitor());
			log.info("Number of tasks started: " + allTaskConfig.size());
			allTaskConfig.stream().forEach(DaemonFactory::addTask);
			//待运行的DSY实例放入队列
			initDsyWaitingInstanceMonitor();
		}
	}
	
	@Override
	public void destroy() throws Exception {
		DaemonFactory.shutdown();
	}
	
	/**
	 * 初始化ETL任务实例
	 */
	private List<DaemonTaskConfig> initEtlTaskMonitor() {
		List<JobInstance> jobInstanceList = jobInstanceService.listAllRunningInstanceList();
		List<DaemonTaskConfig> configList = new ArrayList<>(jobInstanceList.size());
		if(CollectionUtil.isNotEmpty(jobInstanceList)) {
			for(JobInstance jobInstance : jobInstanceList) {
				configList.add(new DaemonTaskConfig(jobInstance.getId().toString(), DaemonTaskType.flink_job));
			}
		}
		return configList;
	}
	
	/**
	 * 初始化【运行中】的数据同步实例
	 * @return
	 */
	private List<DaemonTaskConfig> initDsyRunningInstanceMonitor() {
		List<DsyJobInstance> jobInstanceList = dsyJobInstanceService.listAllRunningInstanceList();
		List<DaemonTaskConfig> configList = new ArrayList<>(jobInstanceList.size());
		if(CollectionUtil.isNotEmpty(jobInstanceList)) {
			for(DsyJobInstance jobInstance : jobInstanceList) {
				configList.add(new DaemonTaskConfig(jobInstance.getId().toString(), DaemonTaskType.datax_job));
			}
		}
		return configList;
	}
	
	/**
	 * 初始化【待执行】的数据同步实例
	 */
	private void initDsyWaitingInstanceMonitor() {
		List<DsyJobInstance> jobInstanceList = dsyJobInstanceService.listAllWaitingInstanceList();
		if(CollectionUtil.isNotEmpty(jobInstanceList)) {
			log.info("Number of DSY:Waiting task pushed queue: " + jobInstanceList.size());
			List<JobRunTarget> targets = jobInstanceList.stream().map(item -> new JobRunTarget(item.getId(), item.getInstanceRunType())).collect(Collectors.toList());
			JobRunInstance.getInstance().add(targets);
			SpringUtil.publishEvent(new JobRunEvent());
		}
	}
}
